@@ -4,10 +4,17 @@ from collections import deque
 from contextvars import ContextVar
 from copy import copy
 
-from .categories import match_category
-from .selector import Element, to_pattern
+from .selector import Element, MatchFunction, to_pattern
 from .selfless import Override, Selfless, choose, override
-from .utils import ABSENT, ACTIVE, COMPLETE, FAILED, call_with_captures
+from .tags import match_tag
+from .utils import (
+    ABSENT,
+    ACTIVE,
+    COMPLETE,
+    FAILED,
+    autocreate,
+    call_with_captures,
+)
 
 _pattern_fit_cache = {}
 
@@ -136,11 +143,20 @@ class Accumulator:
             leaf.status = FAILED
 
     def varset(self, element, varname, category, value):
-        if element.value is ABSENT or element.value == value:
+        if (
+            element.value is ABSENT
+            or element.value == value
+            or (
+                isinstance(element.value, MatchFunction)
+                and element.value.fn(value)
+            )
+        ):
             acc = self.fork(pattern=element) if element.focus else self
             cap = acc.getcap(element)
             if cap:
                 cap.acquire(varname, value)
+            if element.focus and self.rulename == "immediate":
+                acc.run_listeners()
         else:
             self.fail()
 
@@ -249,7 +265,7 @@ def dict_to_collection(*rulesets):
 def check_element(el, name, category):
     if el.name is not None and el.name != name:
         return False
-    elif not match_category(el.category, category):
+    elif not match_tag(el.category, category):
         return False
     else:
         return True
@@ -261,9 +277,9 @@ def fits_pattern(pfn, pattern):
         fcat = None
         fvars = {}
     else:
-        fname = pfn.fn.__name__
+        fname = pfn.origin
         fcat = pfn.fn.__annotations__.get("return", None)
-        fvars = pfn.state_obj.__annotations__
+        fvars = pfn.state_obj.__info__
 
     if not check_element(pattern.element, fname, fcat):
         return False
@@ -274,8 +290,8 @@ def fits_pattern(pfn, pattern):
         if cap.name is None:
             varnames = [
                 var
-                for var, ann in fvars.items()
-                if check_element(cap, var, ann)
+                for var, info in fvars.items()
+                if check_element(cap, var, info["annotation"])
             ]
             if not varnames:
                 return False
@@ -348,7 +364,7 @@ class proceed:
         self.frame.exit()
 
 
-class overlay:
+class BaseOverlay:
     def __init__(self, *rulesets):
         self.rulesets = [rules for rules in rulesets if rules]
 
@@ -421,14 +437,47 @@ def interact(sym, key, category, __self__, value):
             return value
 
 
+class PluginWrapper:
+    def __init__(self, plugin, fn):
+        self.hasoutput = plugin.hasoutput
+        self.plugin = plugin
+        self.fn = fn
+
+    def instantiate(self):
+        return ActivePluginWrapper(self.plugin.instantiate(), self.fn)
+
+
+class ActivePluginWrapper:
+    def __init__(self, active_plugin, fn):
+        self.active_plugin = active_plugin
+        self.fn = fn
+
+    def rules(self):
+        return self.active_plugin.rules()
+
+    def finalize(self):
+        rval = self.active_plugin.finalize()
+        return self.fn(rval)
+
+
 class Collector:
-    def __init__(self, pattern, finalize=None):
+    def __init__(self, pattern, mapper=None, immediate=False):
         self.data = []
         self.pattern = pattern
-        self.finalizer = finalize
+        self.mapper = mapper
+        self.rulename = "immediate" if immediate else "listeners"
 
-        def listener(**kwargs):
-            self.data.append(kwargs)
+        if self.mapper:
+
+            def listener(**kwargs):
+                result = self.mapper(**kwargs)
+                if result is not None:
+                    self.data.append(result)
+
+        else:
+
+            def listener(**kwargs):
+                self.data.append(kwargs)
 
         listener._ptera_argspec = set(self.pattern.all_captures())
         self._listener = listener
@@ -480,11 +529,11 @@ class Collector:
         )
 
     def rules(self):
-        return {self.pattern: {"listeners": [self._listener]}}
+        return {self.pattern: {self.rulename: [self._listener]}}
 
     def finalize(self):
-        if self.finalizer:
-            return self.finalizer(self)
+        if self.mapper:
+            return self.data
         else:
             return self
 
@@ -492,23 +541,22 @@ class Collector:
 class Tap:
     hasoutput = True
 
-    def __init__(self, selector, finalize=None):
+    def __init__(self, selector, mapper=None, immediate=False):
         self.selector = to_pattern(selector)
-        self.finalize = finalize
+        self.mapper = mapper
+        self.immediate = immediate
 
-    def hook(self, finalize):
-        self.finalize = finalize
+    def hook(self, mapper):
+        self.mapper = mapper
         return self
 
     def instantiate(self):
-        return Collector(self.selector, self.finalize)
+        return Collector(
+            self.selector, mapper=self.mapper, immediate=self.immediate
+        )
 
 
 class CallResults:
-    def __init__(self, value):
-        self.value = value
-        setattr(self, "0", self.value)
-
     def __getitem__(self, item):
         if isinstance(item, int):
             item = str(item)
@@ -517,6 +565,9 @@ class CallResults:
         except AttributeError:
             raise IndexError(item)
 
+    def __setitem__(self, item, value):
+        setattr(self, str(item), value)
+
 
 class StateOverlay:
     hasoutput = False
@@ -534,8 +585,8 @@ class StateOverlay:
         return self
 
 
-def _to_plugin(spec):
-    return Tap(spec) if isinstance(spec, str) else spec
+def _to_plugin(spec, **kwargs):
+    return Tap(spec, **kwargs) if isinstance(spec, str) else spec
 
 
 def _collect_plugins(plugins, kwplugins):
@@ -545,38 +596,26 @@ def _collect_plugins(plugins, kwplugins):
     return plugins, any(p.hasoutput for name, p in plugins.items())
 
 
-class PteraFunction(Selfless):
-    def __init__(
-        self, fn, state, callkey=None, plugins=None, return_object=False
-    ):
-        super().__init__(fn, state)
-        self.callkey = callkey
-        self.plugins = plugins or {}
-        self.return_object = return_object
-        self._match_cache = {}
+class Overlay:
+    def __init__(self, plugins={}):
+        self.plugins = dict(plugins)
 
-    def clone(self, **kwargs):
-        self.ensure_state()
-        kwargs = {
-            "fn": self.fn,
-            "state": copy(self.state_obj),
-            "callkey": self.callkey,
-            "plugins": self.plugins,
-            "return_object": self.return_object,
-            **kwargs,
-        }
-        return type(self)(**kwargs)
+    @property
+    def hasoutput(self):
+        return any(p.hasoutput for name, p in self.plugins.items())
 
-    def __getitem__(self, callkey):
-        assert self.callkey is None
-        return self.clone(callkey=callkey)
+    def use(self, *plugins, **kwplugins):
+        plugins, _ = _collect_plugins(plugins, kwplugins)
+        self.plugins.update(plugins)
+        return self
 
     def tweak(self, values, priority=2):
         values = {
             to_pattern(k): lambda __v=v, **_: override(__v, priority)
             for k, v in values.items()
         }
-        return self.using(StateOverlay(values))
+        self.plugins[f"#{len(self.plugins)}"] = StateOverlay(values)
+        return self
 
     def rewrite(self, values, full=False, priority=2):
         def _wrapfn(fn, full=True):
@@ -590,60 +629,158 @@ class PteraFunction(Selfless):
             return newfn
 
         values = {k: _wrapfn(v, full=full) for k, v in values.items()}
-        return self.using(StateOverlay(values))
-
-    def using(self, *plugins, **kwplugins):
-        plugins, return_object = _collect_plugins(plugins, kwplugins)
-        return self.clone(
-            plugins={**self.plugins, **plugins}, return_object=return_object,
-        )
-
-    def use(self, *plugins, **kwplugins):
-        plugins, _ = _collect_plugins(plugins, kwplugins)
-        self.plugins.update(plugins)
+        self.plugins[f"#{len(self.plugins)}"] = StateOverlay(values)
         return self
 
+    @autocreate
+    def using(self, *plugins, **kwplugins):
+        plugins, _ = _collect_plugins(plugins, kwplugins)
+        return Overlay(plugins={**self.plugins, **plugins})
+
+    @autocreate
+    def tweaking(self, values, priority=2):
+        ol = Overlay(self.plugins)
+        return ol.tweak(values, priority=priority)
+
+    @autocreate
+    def rewriting(self, values, full=False, priority=2):
+        ol = Overlay(self.plugins)
+        return ol.rewrite(values, full=full, priority=priority)
+
     def collect(self, query):
         plugin = _to_plugin(query)
 
         def deco(fn):
-            self.plugins[fn.__name__] = plugin.hook(fn)
+            self.plugins[fn.__name__] = PluginWrapper(plugin, fn)
 
         return deco
 
-    def on(self, query, full=False, all=False):
-        plugin = _to_plugin(query)
+    def on(self, query, full=False, all=False, immediate=True):
+        plugin = _to_plugin(query, immediate=immediate)
 
         def deco(fn):
-            def finalize(coll):
-                if full:
-                    return coll.map_full(fn)
-                elif all:
-                    return coll.map_all(fn)
-                else:
-                    return coll.map(fn)
+            def mapper(**kwargs):
+                if all:
+                    kwargs = {key: cap.values for key, cap in kwargs.items()}
+                elif not full:
+                    kwargs = {key: cap.value for key, cap in kwargs.items()}
+                return fn(**kwargs)
 
-            self.plugins[fn.__name__] = plugin.hook(finalize)
+            self.plugins[fn.__name__] = plugin.hook(mapper)
 
         return deco
 
-    def __call__(self, *args, **kwargs):
-        self.ensure_state()
+    def __enter__(self):
         rulesets = []
         plugins = {name: p.instantiate() for name, p in self.plugins.items()}
         for plugin in plugins.values():
             rulesets.append(plugin.rules())
-        with overlay(*rulesets):
-            with proceed(self):
-                if self.callkey is not None:
-                    interact("#key", None, None, self, self.callkey)
-                rval = super().__call__(*args, **kwargs)
+        self._ol = BaseOverlay(*rulesets)
+        self._ol.__enter__()
+        self._results = CallResults()
+        self._inst_plugins = plugins
+        return self._results
 
-        callres = CallResults(rval)
-        for name, plugin in plugins.items():
-            setattr(callres, name, plugin.finalize())
+    def __exit__(self, typ, exc, tb):
+        self._ol.__exit__(None, None, None)
+        for name, plugin in self._inst_plugins.items():
+            setattr(self._results, name, plugin.finalize())
+
+
+class PteraFunction(Selfless):
+    def __init__(
+        self,
+        fn,
+        state,
+        overlay=None,
+        return_object=False,
+        origin=None,
+        attachments=None,
+        partial_args=(),
+    ):
+        super().__init__(fn, state)
+        self.overlay = overlay or Overlay()
+        self.return_object = return_object
+        self.origin = origin or self
+        self.partial_args = partial_args
+        self.attachments = attachments or {}
+
+    def clone(self, **kwargs):
+        self.ensure_state()
+        kwargs = {
+            "fn": self.fn,
+            "state": copy(self.state_obj),
+            "overlay": Overlay(self.overlay.plugins),
+            "return_object": self.return_object,
+            "origin": self.origin,
+            "partial_args": self.partial_args,
+            "attachments": self.attachments,
+            **kwargs,
+        }
+        return type(self)(**kwargs)
+
+    def attach(self, **values):
+        return self.clone(attachments={**self.attachments, **values})
+
+    def __getitem__(self, callkey):
+        assert isinstance(callkey, list)
+        if len(callkey) == 1:
+            (callkey,) = callkey
+        attach = {"key": callkey}
+        if isinstance(callkey, list):
+            for i, v in enumerate(callkey):
+                attach[f"key{i}"] = v
+        return self.attach(**attach)
+
+    def use(self, *plugins, **kwplugins):
+        self.overlay.use(*plugins, **kwplugins)
+        return self
+
+    def tweak(self, values, priority=2):
+        self.overlay.tweak(values, priority=priority)
+        return self
+
+    def rewrite(self, values, full=False, priority=2):
+        self.overlay.rewrite(values, full=full, priority=priority)
+        return self
+
+    def tweaking(self, values, priority=2):
+        return self.clone(
+            overlay=self.overlay.tweaking(values, priority=priority)
+        )
+
+    def rewriting(self, values, full=False, priority=2):
+        return self.clone(
+            overlay=self.overlay.rewriting(values, full=full, priority=priority)
+        )
+
+    def using(self, *plugins, **kwplugins):
+        ol = self.overlay.using(*plugins, **kwplugins)
+        return self.clone(overlay=ol, return_object=ol.hasoutput)
+
+    def collect(self, query):
+        return self.overlay.collect(query)
+
+    def on(self, query, full=False, all=False):
+        return self.overlay.on(query, full=full, all=all)
+
+    def __get__(self, obj, typ):
+        if obj is None:
+            return self
+        else:
+            return self.clone(partial_args=self.partial_args + (obj,))
+
+    def __call__(self, *args, **kwargs):
+        self.ensure_state()
+        with self.overlay as callres:
+            with proceed(self):
+                if self.attachments:
+                    for k, v in self.attachments.items():
+                        interact(f"#{k}", None, None, self, v)
+                rval = super().__call__(*self.partial_args, *args, **kwargs)
+                callres["0"] = callres["value"] = rval
 
         if self.return_object:
             return callres
         else:
-            return rval
+            return callres.value
