@@ -3,10 +3,12 @@
 
 import ast
 import builtins
+import re
 import sys
+from dataclasses import dataclass
 
 from . import opparse
-from .categories import Category
+from .tags import Tag
 from .utils import ABSENT
 
 
@@ -221,7 +223,7 @@ class Call(metaclass=InternedMC):
             enc = child.encode()
             enc = f"> {enc}" if child.immediate else f">> {enc}"
             caps.append(enc)
-        caps = "" if not caps else "{" + ", ".join(caps) + "}"
+        caps = "" if not caps else "(" + ", ".join(caps) + ")"
         return f"{name}{caps}"
 
     def __str__(self):
@@ -233,30 +235,33 @@ class Call(metaclass=InternedMC):
 parser = opparse.Parser(
     lexer=opparse.Lexer(
         {
-            r"\s*(?:\bas\b|>>|!+|[(){}\[\]>:,$=])?\s*": "OPERATOR",
+            r"\s*(?:\bas\b|>>|!+|\[\[|\]\]|[(){}\[\]>:,$=~])?\s*": "OPERATOR",
             r"[a-zA-Z_0-9#*.]+": "WORD",
+            r"'[^']*'": "STRING",
         }
     ),
     order=opparse.OperatorPrecedenceTower(
         {
             ",": opparse.rassoc(10),
             ("", ">", ">>"): opparse.rassoc(100),
-            "=": opparse.lassoc(120),
+            ("=", "~"): opparse.lassoc(120),
             ("!", "!!"): opparse.lassoc(150),
             ":": opparse.lassoc(300),
             "as": opparse.rassoc(350),
             "$": opparse.lassoc(400),
-            ("(", "[", "{"): opparse.obrack(200),
-            (")", "]", "}"): opparse.cbrack(500),
+            ("(", "[", "{", "[["): opparse.obrack(200),
+            (")", "]", "}", "]]"): opparse.cbrack(500),
             ": WORD": opparse.lassoc(1000),
+            ": STRING": opparse.lassoc(1000),
         }
     ),
 )
 
 
-def _guarantee_call(parent, context):
+def _guarantee_call(parent, context, resolve=True):
     if isinstance(parent, Element):
-        parent = parent.clone(capture=None).without_focus()
+        name = VSymbol(parent.name) if parent.name and resolve else parent.name
+        parent = parent.clone(capture=None, name=name).without_focus()
         immediate = context == "incall"
         parent = Call(element=parent, captures=(), immediate=immediate)
     assert isinstance(parent, Call)
@@ -353,19 +358,14 @@ def make_nested_pfx(node, _, child, context):
         return child.clone(immediate=False)
 
 
-@evaluate.register_action("X : X")
-def make_class(node, element, klass, context):
-    element = evaluate(element, context=context)
-    klass = evaluate(klass, context=context)
-    assert isinstance(klass, Element)
-    assert not element.category
-    return element.clone(category=klass.name)
-
-
 @evaluate.register_action("_ : X")
-def make_class_prefix(node, _, klass, context):
-    klass = evaluate(klass, context=context)
-    return Element(name=None, category=klass.name, capture=None,)
+@evaluate.register_action("X : X")
+def make_class(node, element, tag, context):
+    element = (
+        evaluate(element, context=context) if element else Element(name=None)
+    )
+    tag = value_evaluate(tag)
+    return element.clone(category=tag)
 
 
 @evaluate.register_action("_ ! X")
@@ -391,33 +391,45 @@ def make_dollar(node, _, name, context):
 
 
 @evaluate.register_action("X [ X ] _")
-def make_instance(node, element, key, _, context):
+def make_index(node, element, key, _, context, resolve_call=False):
+    def _make_key(key, suffix):
+        assert isinstance(key, Element)
+        if key.value is not ABSENT:
+            assert key.name is None
+            val = key.value
+        else:
+            val = VSymbol(key.name) if key.name is not None else ABSENT
+        return Element(
+            name=f"#key{suffix}",
+            value=val,
+            category=key.category,
+            capture=key.capture if key.name != key.capture else None,
+            key_field="value" if key.name is None else None,
+        )
+
     element = evaluate(element, context=context)
     key = evaluate(key, context=context)
     assert isinstance(element, Element)
-    assert isinstance(key, Element)
-    element = _guarantee_call(element, context=context)
-    key = Element(
-        name="#key",
-        value=key.name if key.name is not None else ABSENT,
-        category=key.category,
-        capture=key.capture if key.name != key.capture else None,
-        key_field="value" if key.name is None else None,
-    )
-    return element.clone(captures=element.captures + (key,))
+    element = _guarantee_call(element, context=context, resolve=resolve_call)
+
+    if isinstance(key, list):
+        keys = tuple(_make_key(key, i) for i, key in enumerate(key))
+    else:
+        keys = (_make_key(key, ""),)
+
+    return element.clone(captures=element.captures + keys)
 
 
-@evaluate.register_action("X { _ } _")
-def make_call_empty_capture(node, fn, _1, _2, context):
+@evaluate.register_action("X [[ X ]] _")
+def make_function_index(node, element, key, _, context):
+    return make_index(node, element, key, _, context, resolve_call=True)
+
+
+@evaluate.register_action("X ( _ ) _")
+@evaluate.register_action("X ( X ) _")
+def make_call_capture(node, fn, names, _, context):
     fn = evaluate(fn, context=context)
-    fn = _guarantee_call(fn, context=context)
-    return fn
-
-
-@evaluate.register_action("X { X } _")
-def make_call_capture(node, fn, names, _2, context):
-    fn = evaluate(fn, context=context)
-    names = evaluate(names, context="incall")
+    names = evaluate(names, context="incall") if names else []
     names = names if isinstance(names, list) else [names]
     fn = _guarantee_call(fn, context=context)
     caps = tuple(name for name in names if isinstance(name, Element))
@@ -455,12 +467,26 @@ def make_as(node, element, name, context):
         return element.clone(captures=element.captures + (new_capture,))
 
 
+@evaluate.register_action("_ = X")
 @evaluate.register_action("X = X")
-def make_equals(node, element, value, context):
-    element = evaluate(element, context=context)
-    value = evaluate(value, context=context)
-    assert isinstance(value, Element)
-    return element.clone(value=value.name, capture=None)
+def make_equals(node, element, value, context, matchfn=False):
+    if element is None:
+        element = Element(name=None)
+    else:
+        element = evaluate(element, context=context)
+    value = value_evaluate(value)
+    if matchfn:
+        capture = element.capture
+        value = VCall(MatchFunction, (value,))
+    else:
+        capture = None
+    return element.clone(value=value, capture=capture)
+
+
+@evaluate.register_action("_ ~ X")
+@evaluate.register_action("X ~ X")
+def make_matchfn(node, element, value, context):
+    return make_equals(node, element, value, context, matchfn=True)
 
 
 @evaluate.register_action("SYMBOL")
@@ -469,12 +495,7 @@ def make_symbol(node, context):
         element = Element(name=None)
     else:
         value = node.value
-        cap = node.value
-        try:
-            value = int(value)
-            cap = None
-        except ValueError:
-            pass
+        cap = node.value[1:] if node.value.startswith("#") else node.value
         focus = context == "root"
         element = Element(
             name=value,
@@ -484,6 +505,106 @@ def make_symbol(node, context):
     return element
 
 
+class VNode:
+    pass
+
+
+@dataclass(frozen=True)
+class VSymbol(VNode):
+    value: str
+
+    def eval(self, env):
+        x = self.value
+
+        if re.match(r"[0-9]+\.[0-9]*", x):
+            return float(x)
+        elif re.match(r"[0-9]+", x):
+            return int(x)
+        elif re.match(r"'[^']*'", x):
+            return x[1:-1]
+
+        start, *parts = x.split(".")
+        if start in env:
+            curr = env[start]
+        elif hasattr(builtins, start):
+            return getattr(builtins, start)
+        else:
+            raise Exception(f"Could not resolve '{start}'.")
+
+        for part in parts:
+            curr = getattr(curr, part)
+
+        return curr
+
+    def __str__(self):
+        return str(self.value)
+
+
+@dataclass(frozen=True)
+class VCall(VNode):
+    fn: object
+    args: tuple
+
+    def eval(self, env):
+        fn = _eval(self.fn, env)
+        args = []
+        kwargs = {}
+        for arg in self.args:
+            if isinstance(arg, VKeyword):
+                kwargs[arg.key.value] = _eval(arg.value, env)
+            else:
+                args.append(_eval(arg, env))
+        return fn(*args, **kwargs)
+
+
+@dataclass(frozen=True)
+class VKeyword(VNode):
+    key: object
+    value: object
+
+
+def _eval(x, env):
+    if isinstance(x, VNode):
+        return x.eval(env)
+    else:
+        return x
+
+
+value_evaluate = Evaluator()
+
+
+@value_evaluate.register_action("X , X")
+def vmake_sequence(node, a, b, context):
+    a = value_evaluate(a)
+    b = value_evaluate(b)
+    if not isinstance(b, list):
+        b = [b]
+    return [a, *b]
+
+
+@value_evaluate.register_action("X ( _ ) _")
+@value_evaluate.register_action("X ( X ) _")
+def vmake_call(node, fn, args, _, context):
+    fn = value_evaluate(fn)
+    args = value_evaluate(args) if args else []
+    if not isinstance(args, list):
+        args = [args]
+    return VCall(fn, tuple(args))
+
+
+@value_evaluate.register_action("X = X")
+def vmake_keyword(node, key, value, context):
+    key = value_evaluate(key)
+    assert isinstance(key, VSymbol)
+    value = value_evaluate(value)
+    return VKeyword(key, value)
+
+
+@value_evaluate.register_action("SYMBOL")
+def vmake_symbol(node, context):
+    return VSymbol(node.value)
+
+
 def parse(x):
     return evaluate(parser(x))
 
@@ -531,20 +652,9 @@ def _find_eval_env(s, fr):
     return ev and ev[1]
 
 
-def _eval(s, env):
-    if not isinstance(s, str):
-        return s
-    start, *parts = s.split(".")
-    if start in env:
-        curr = env[start]
-    elif hasattr(builtins, start):
-        return getattr(builtins, start)
-    else:
-        raise Exception(f"Could not resolve '{start}'.")
-
-    for part in parts:
-        curr = getattr(curr, part)
-    return curr
+@dataclass(frozen=True)
+class MatchFunction:
+    fn: object
 
 
 def _resolve(pattern, env):
@@ -556,10 +666,12 @@ def _resolve(pattern, env):
             children=tuple(_resolve(x, env) for x in pattern.children),
         )
     elif isinstance(pattern, Element):
+        name = _eval(pattern.name, env)
         category = _eval(pattern.category, env)
-        if category is not None and not isinstance(category, Category):
-            raise TypeError(f"A pattern can only be a Category.")
-        return pattern.clone(category=category)
+        value = _eval(pattern.value, env)
+        if category is not None and not isinstance(category, Tag):
+            raise TypeError(f"A pattern can only be a Tag.")
+        return pattern.clone(name=name, category=category, value=value)
 
 
 def _to_pattern(pattern, context="root"):
